import io
import random
import sys

import matplotlib.pyplot as plt
import numpy as np
import pygame as pg
import seaborn as sns
import torch as th
from PIL import Image

from mu_alpha_zero.General.az_game import AlphaZeroGame


class TicTacToeGameManager(AlphaZeroGame):
    """
    This class is the game manager for the game of Tic Tac Toe and its variants.
    """

    def __init__(self, board_size: int, headless: bool, num_to_win=None) -> None:
        # TODO: Implement the possibility to play over internet using sockets.
        self.player = 1
        self.enemy_player = -1
        self.board_size = board_size
        self.board = self.initialize_board()
        self.headless = headless
        self.num_to_win = self.init_num_to_win(num_to_win)
        self.screen = self.create_screen(headless)

    def play(self, player: int, index: tuple) -> None:
        self.board[index] = player

    def init_num_to_win(self, num_to_win: int | None) -> int:
        if num_to_win is None:
            num_to_win = self.board_size
        if num_to_win > self.board_size:
            raise Exception("Num to win can't be greater than board size")
        return num_to_win

    def initialize_board(self):
        board = np.zeros((self.board_size, self.board_size), dtype=np.int8)
        return board

    def get_random_valid_action(self, observations: np.ndarray) -> list:
        valid_moves = self.get_valid_moves(observations)
        if len(valid_moves) == 0:
            raise Exception("No valid moves")
        return random.choice(valid_moves)

    def create_screen(self, headless):
        if headless:
            return

        pg.init()
        pg.display.set_caption("Tic Tac Toe")
        board_rect_size = self.board_size * 100
        screen = pg.display.set_mode((board_rect_size, board_rect_size))
        return screen

    def full_iterate_array(self, arr: np.ndarray, func: callable) -> list:
        """
        This function iterates over all rows, columns and the two main diagonals of supplied array,
        applies the supplied function to each of them and returns the results in a list.
        :param arr: a 2D numpy array.
        :param func: a callable function that takes a 1D array as input and returns a result.
        :return: A list of results.
        """
        results = []
        results.append("row")
        for row in arr:
            results.append(func(row.reshape(-1)))

        results.append("col")
        for col in arr.T:
            results.append(func(col.reshape(-1)))

        diags = [np.diag(arr, k=i) for i in range(-arr.shape[0] + 1, arr.shape[1])]
        flipped_diags = [np.diag(np.fliplr(arr), k=i) for i in range(-arr.shape[0] + 1, arr.shape[1])]
        diags.extend(flipped_diags)
        for idx, diag in enumerate(diags):
            if idx == 0:
                results.append("diag_left")
            elif idx == len(diags) // 2:
                results.append("diag_right")
            results.append(func(diag.reshape(-1)))

        # results.append("diag_left")
        # results.append(func(arr.diagonal().reshape(-1)))
        # results.append("diag_right")
        # results.append(func(np.fliplr(arr).diagonal().reshape(-1)))
        return results

    def eval_board(self, board: np.ndarray,check_end: bool = True) -> int | None:
        if self.is_board_full(board):
            return 0
        score = 0
        for i in range(self.num_to_win, self.num_to_win // 2, -1):
            current_won = self.check_partial_win(-1, self.num_to_win, board=board)
            opp_won = self.check_partial_win(1, self.num_to_win, board=board)
            if current_won:
                score += 1 * 2
            if opp_won:
                score -= 1
        if check_end:
            return None if self.game_result(-1, board) is None else score
        return score

    def full_iterate_array_all_diags(self, arr: np.ndarray, func: callable):

        results = []
        for row in arr:
            results.append(func(row.reshape(-1)))

        for col in arr.T:
            results.append(func(col.reshape(-1)))

        diags = [np.diag(arr, k=i) for i in range(-arr.shape[0] + 1, arr.shape[1])]
        flipped_diags = [np.diag(np.fliplr(arr), k=i) for i in range(-arr.shape[0] + 1, arr.shape[1])]
        diags.extend(flipped_diags)
        for diag in diags:
            # if diag.size < self.num_to_win:
            #     continue
            results.append(func(diag.reshape(-1)))

        return results

    def extract_all_vectors(self, board: np.ndarray):
        vectors = board.tolist() + board.T.tolist()
        vectors.extend([np.diag(board, k=i) for i in range(-board.shape[0] + 1, board.shape[1])])
        vectors.extend([np.diag(np.fliplr(board), k=i) for i in range(-board.shape[0] + 1, board.shape[1])])
        # pad vectors with -3's to have the same length as the board size and stack them
        vectors = [np.pad(vector, (0, self.board_size - len(vector)), constant_values=-3) for vector in vectors]
        return np.array(vectors)

    def check_win(self, player: int, board=None) -> bool:
        # if self.num_to_win == self.board_size:
        #     return self.check_full_win(player, board=board)
        # else:
        return self.check_partial_win(player, self.num_to_win, board=board)

    def check_full_win(self, player: int, board=None) -> bool:
        """
        This function checks if the supplied player has won the game with a full win (num_to_win == board_size).
        :param player: The player to check for (1 or -1).
        :param board: The board to check on. If None, the current board is used.
        :return: True if the player has won, False otherwise.
        """
        if board is None:
            board = self.get_board()
        matches = self.full_iterate_array(board, lambda part: np.all(part == player))
        for match in matches:
            if not isinstance(match, str) and match:
                return True

        return False

    def check_partial_win(self, player: int, n: int, board=None) -> bool:
        """
        This function checks if the supplied player has won the game with a partial win (num_to_win < board_size).

        :param player: The player to check for (1 or -1).
        :param n: The number of consecutive pieces needed to win.
        :param board: The board to check on. If None, the current board is used.
        :return: True if the player has won, False otherwise.
        """
        if board is None:
            board = self.get_board()
        matches = self.full_iterate_array_all_diags(board,
                                                    lambda part:
                                                    np.convolve((part == player), np.ones(n, dtype=np.int8),
                                                                "valid"))

        for match in matches:
            if np.any(match == n):
                return True

        return False

    def check_partial_win_vectorized(self, player: int, n: int, board=None) -> bool:
        if board is None:
            board = self.get_board()
        vectors = th.tensor(self.extract_all_vectors(board)).unsqueeze(0)
        weight = th.ones((1, 1, 1, n), dtype=th.long)
        vectors_where_player = th.where(vectors == player, 1, 0).long()
        res = th.nn.functional.conv2d(vectors_where_player, weight=weight)
        return th.any(res == n).item()

    def check_partial_win_to_index(self, player: int, n: int, board=None) -> dict[tuple, str] | dict:
        """
        This variation of check_partial_win returns the index of the first partial win found.
        The index being the index of the first piece in the winning sequence.
        :param player: The player to check for (1 or -1).
        :param n: The number of consecutive pieces needed to win.
        :param board: The board to check on. If None, the current board is used.
        :return: A dictionary containing the index and the position of the winning sequence.
        """
        if board is None:
            board = self.get_board()
        indices = self.full_iterate_array(board,
                                          lambda part:
                                          np.convolve((part == player), np.ones(n, dtype=int),
                                                      "valid"))
        indices = [x.tolist() if not isinstance(x, str) else x for x in indices]

        pos = "row"
        for vector in indices:
            if isinstance(vector, str):
                pos = vector
                continue
            for el in vector:
                if el == n:
                    vector_index = indices.index(vector)
                    pos_index = indices.index(pos)
                    # num_strings_before = len([x for index,x in enumerate(indices) if isinstance(x,str) and index < pos_index])
                    element_index = vector_index - (pos_index + 1) if vector_index > pos_index else indices.index(
                        vector, pos_index)
                    diag_idx = {
                        "index": ((self.board_size - 1) - element_index, 0)} if element_index < self.board_size else {
                        "index": (0, element_index - self.board_size)}
                    diag_idx["pos"] = pos
                    match pos:
                        case "row":
                            return {"index": (element_index, 0), "pos": pos}
                        case "col":
                            return {"index": (0, element_index), "pos": pos}
                        case "diag_left":
                            return diag_idx
                        case "diag_right":
                            return diag_idx

                        # case "diag_right":

        return {"index": None, "pos": None}

    def return_index_if_valid(self, index: tuple, return_on_fail: tuple = ()) -> tuple:
        if index[0] < 0 or index[0] >= self.board_size or index[1] < 0 or index[1] >= self.board_size:
            return return_on_fail
        return index

    def make_fresh_instance(self):
        return TicTacToeGameManager(self.board_size, self.headless, num_to_win=self.num_to_win)

    def get_previous(self, index: tuple, pos: str, n: int):
        if np.array(index).all() == 0:
            return index
        match pos:
            case "row":
                return self.return_index_if_valid((index[0], index[1] - n), return_on_fail=index)
            case "col":
                return self.return_index_if_valid((index[0] - n, index[1]), return_on_fail=index)
            case "diag_left":
                return self.return_index_if_valid((index[0] - n, index[1] - n), return_on_fail=index)
            case "diag_right":
                return self.return_index_if_valid((index[0] - n, index[1] + n), return_on_fail=index)

    def get_next(self, index: tuple, pos: str, n: int):
        if np.array(index).all() == self.board_size - 1:
            return index
        match pos:
            case "row":
                return self.return_index_if_valid((index[0], index[1] + n), return_on_fail=index)
            case "col":
                return self.return_index_if_valid((index[0] + n, index[1]), return_on_fail=index)
            case "diag_left":
                return self.return_index_if_valid((index[0] + n, index[1] + n), return_on_fail=index)
            case "diag_right":
                return self.return_index_if_valid((index[0] + n, index[1] - n), return_on_fail=index)

    def is_board_full(self, board=None) -> bool:
        if board is None:
            board = self.get_board()
        return np.all(board != 0)

    def get_board(self):
        return self.board.copy()

    def reset(self):
        self.board = self.initialize_board()
        return self.board.copy()

    def get_board_size(self):
        return self.board_size

    def render(self) -> bool:
        if self.headless:
            return False

        self.screen.fill((0, 0, 0))
        self._draw_board()
        for row in range(self.board_size):
            for col in range(self.board_size):
                if self.board[row][col] == self.player:
                    self._draw_circle(col * 100 + 50, row * 100 + 50)

                elif self.board[row][col] == self.enemy_player:
                    self._draw_cross(col * 100 + 50, row * 100 + 50)

        pg.event.pump()
        pg.display.flip()
        return True

    def _draw_circle(self, x, y) -> None:
        if self.headless:
            return
        pg.draw.circle(self.screen, "green", (x, y), 40, 1)

    def _draw_cross(self, x, y) -> None:
        if self.headless:
            return
        pg.draw.line(self.screen, "red", (x - 40, y - 40), (x + 40, y + 40), 1)
        pg.draw.line(self.screen, "red", (x + 40, y - 40), (x - 40, y + 40), 1)

    def _draw_board(self):
        for x in range(0, self.board_size * 100, 100):
            for y in range(0, self.board_size * 100, 100):
                pg.draw.rect(self.screen, (255, 255, 255), pg.Rect(x, y, 100, 100), 1)

    def is_empty(self, index: tuple) -> bool:
        return self.board[index] == 0

    def get_valid_moves(self, observation: np.ndarray, player: int or None = None) -> list:
        """
        Legal moves are the empty spaces on the board.
        :param observation: A 2D numpy array representing the current state of the game.
        :param player: The player to check for. Since the game is symmetric, this is ignored.
        :return: A list of legal moves.
        """
        legal_moves = []
        observation = observation.reshape(self.board_size, self.board_size)
        for row in range(self.board_size):
            for col in range(self.board_size):
                if observation[row][col] == 0:
                    legal_moves.append([row, col])
        return legal_moves

    def pygame_quit(self) -> bool:
        if self.headless:
            return False
        pg.quit()
        return True

    def get_click_coords(self):
        if self.headless:
            return
        mouse_pos = (x // 100 for x in pg.mouse.get_pos())
        if pg.mouse.get_pressed()[0]:  # left click
            return mouse_pos

    def get_human_input(self, board: np.ndarray):
        if self.headless:
            return
        while True:
            self.check_pg_events()
            if self.get_click_coords() is not None:
                x, y = self.get_click_coords()
                if board[y][x] == 0:
                    return y, x

            # time.sleep(1 / 60)

    def check_pg_events(self):
        if self.headless:
            return
        for event in pg.event.get():
            if event.type == pg.QUIT:
                self.pygame_quit()
                sys.exit(0)

    def network_to_board(self, move):
        """
        Converts an integer move from the network to a board index.
        :param move: An integer move selected from the network probabilities.
        :return: A tuple representing the board index (int,int).
        """
        return np.unravel_index(move, self.board.shape)

    def save_screenshot_with_probabilities(self, action_probs, path):
        if self.headless:
            return
        plt.figure(figsize=(15, 10))
        labels, probabilities = zip(*action_probs.items())
        print(probabilities)
        labels = [f"{np.unravel_index(x, self.board.shape)[0]};{np.unravel_index(x, self.board.shape)[1]}" for x in
                  labels]
        sns.barplot(x=labels, y=probabilities)
        plt.xticks(rotation=90)
        plt.xlabel("Move")
        plt.ylabel("Probability")
        plt.title("Action probabilities")
        # Save the plot to a buffer
        buf = io.BytesIO()
        plt.savefig(buf, format='png', bbox_inches="tight")
        buf.seek(0)
        plot_img = Image.open(buf)

        # Convert the pg surface to an image
        surface_buffer = pg.image.tostring(self.screen, 'RGBA')
        surface_img = Image.frombytes('RGBA', self.screen.get_size(), surface_buffer)

        # Concatenate the images vertically
        total_height = plot_img.height + surface_img.height
        combined_img = Image.new('RGB', (max(plot_img.width, surface_img.width), total_height))
        combined_img.paste(plot_img, (0, 0))
        combined_img.paste(surface_img, (0, plot_img.height))
        combined_img.save(path)

    @staticmethod
    def get_canonical_form(board, player) -> np.ndarray:
        return board * player

    def get_next_state(self, board: np.ndarray, action: int or tuple, player: int) -> np.ndarray:
        if isinstance(action, int):
            action = self.network_to_board(action)
        board_ = board.copy()
        board_[action] = player
        return board_

    def set_headless(self, val: bool):
        self.headless = val

    def __str__(self):
        return str(self.board).replace('1', 'X').replace('-1', 'O')
#
# if __name__ == "__main__":
#     sample_arr = np.array([[-1,1,-1,1,1],[-1,1,1,-1,-1],[1,-1,1,1,0],[-1,1,1,-1,0],[0,0,-1,1,1]])
#     game_manager = GameManager(5, True,num_to_win=3)
# res = game_manager.check_partial_win_vectorized(1,3,sample_arr)
# res = game_manager.game_result(-1,sample_arr)
# print(res)
# print(game_manager.check_partial_win(1,3,sample_arr))



import math

import numpy as np
import optuna
import torch as th
from PIL import Image

from mu_alpha_zero.config import MuZeroConfig


def add_actions_to_obs(observations: th.Tensor, actions: th.Tensor, dim=0):
    return th.cat((observations, actions), dim=dim)


def match_action_with_obs(observations: th.Tensor, action: int):
    action = th.full((1, observations.shape[1], observations.shape[2]), action, dtype=th.float32,
                     device=observations.device)
    return add_actions_to_obs(observations, action)


def match_action_with_obs_batch(observation_batch: th.Tensor, action_batch: list[int]):
    tensors = [th.full((1, 1, observation_batch.shape[2], observation_batch.shape[3]), action,
                       dtype=th.float32, device=observation_batch.device) for action in action_batch]
    actions = th.cat(tensors, dim=0)
    return add_actions_to_obs(observation_batch, actions, dim=1)


def resize_obs(observations: np.ndarray, size: tuple[int, int]) -> np.ndarray:
    obs = Image.fromarray(observations)
    obs = obs.resize(size)
    return np.array(obs)


def scale_state(state: np.ndarray):
    # scales the given state to be between 0 and 1
    return state / 255


def scale_action(action: int, num_actions: int):
    return action / (num_actions - 1)


def scale_reward_value(value: th.Tensor, e: float = 0.001):
    if isinstance(value, float) or isinstance(value, np.float32):
        scaled_v = np.sign(value) * (np.sqrt(np.abs(value) + 1) - 1 + value * e)
        return np.array([scaled_v])
    return th.sign(value) * (th.sqrt(th.abs(value) + 1) - 1 + value * e)


def scale_reward(reward: float):
    return math.log(reward + 1, 5)


def mz_optuna_parameter_search(n_trials: int, init_net_path: str, storage: str or None, study_name: str, game,
                               muzero_config: MuZeroConfig, in_memory: bool = False):
    def objective(trial):
        num_mc_simulations = trial.suggest_int("num_mc_simulations", 100, 1200)
        num_self_play_games = trial.suggest_int("num_self_play_games", 100, 500)
        num_epochs = trial.suggest_int("num_epochs", 100, 3000)
        lr = trial.suggest_float("lr", 1e-8, 1e-2, log=True)
        tau = trial.suggest_float("tau", 0.5, 1.5)
        arena_tau = trial.suggest_float("arena_tau", 0.01, 0.5)
        c = trial.suggest_float("c", 0.5, 5)
        c2 = trial.suggest_categorical("c2", [19652, 10_000, 0.01, 10, 0.1])
        K = trial.suggest_int("K", 1, 10)

        muzero_config.num_simulations = num_mc_simulations
        muzero_config.self_play_games = num_self_play_games
        muzero_config.epochs = num_epochs
        muzero_config.lr = lr
        muzero_config.tau = tau
        muzero_config.c = c
        muzero_config.c2 = c2
        muzero_config.arena_tau = arena_tau
        muzero_config.K = K
        muzero_config.num_iters = 5

        device = th.device("cuda" if th.cuda.is_available() else "cpu")
        trial.net_action_size = int(game.get_num_actions())
        network = MuZeroNet.make_from_config(muzero_config).to(device)
        network.load_state_dict(th.load(init_net_path))
        tree = MuZeroSearchTree(game.make_fresh_instance(), muzero_config)
        net_player = NetPlayer(game.make_fresh_instance(),
                               **{"network": network, "monte_carlo_tree_search": tree})
        arena = MzArena(game.make_fresh_instance(), muzero_config, device)
        trainer = Trainer.create(muzero_config, game.make_fresh_instance(), network, tree, net_player,
                                 headless=True, arena_override=arena, checkpointer_verbose=False)
        mean = trainer.self_play_and_train()
        trial.report(mean, trial.number)
        print(f"Trial {trial.number} finished with win freq {mean}.")
        del trainer
        del network
        del tree
        del net_player
        return mean

    from mu_alpha_zero.MuZero.MZ_Arena.arena import MzArena
    from mu_alpha_zero.MuZero.MZ_MCTS.mz_search_tree import MuZeroSearchTree
    from mu_alpha_zero.MuZero.Network.networks import MuZeroNet
    from mu_alpha_zero.AlphaZero.Network.trainer import Trainer
    from mu_alpha_zero.AlphaZero.Arena.players import NetPlayer

    muzero_config.show_tqdm = False
    if in_memory:
        study = optuna.create_study(study_name=study_name)
    else:
        if storage is None:
            raise ValueError("Storage can't be None if in_memory is False.")
        study = optuna.load_study(study_name=study_name, storage=storage)
    study.optimize(objective, n_trials=n_trials)



import os
import pickle

from portalocker import Lock


class DataPickler:
    """
    Class for efficiently storing and retrieving data from the file system.
    """

    def __init__(self, pickle_dir: str):
        self.processed_count = 0
        self.pickle_dir = self.__init_pickle_dir(pickle_dir)

    def __init_pickle_dir(self, pickle_dir: str) -> str:
        os.makedirs(pickle_dir, exist_ok=True)
        return pickle_dir

    def pickle_buffer(self, buffer: list):
        for index in range(len(buffer[0])):
            data = [item[index] for item in buffer]
            with Lock(f"{self.pickle_dir}/item_{index}.pkl", "ab", timeout=90) as f:
                pickle.dump(data, f)
                f.flush()
                os.fsync(f.fileno())
        self.processed_count += 1

    def load_index(self, index: int):
        file = f"{self.pickle_dir}/item_{index}.pkl"
        data = []
        with Lock(file, "rb", timeout=90) as f:
            while True:
                try:
                    data.extend(pickle.load(f))
                except EOFError:
                    break

        return data

    def load_all(self, batch_size: int, indexes: list, K: int = 1):
        files = [x for x in os.listdir(self.pickle_dir) if x.endswith(".pkl")]
        files.sort(key=lambda x: int(x.split("_")[1].split(".")[0]))
        data = [[] for _ in range(len(files))]
        for i, file in enumerate(files):
            with Lock(f"{self.pickle_dir}/{file}", "rb", timeout=90) as f:
                last_len = 0
                while True:
                    if len(data[i]) >= batch_size:
                        break
                    try:
                        file_data = pickle.load(f)
                        for index in indexes:
                            if last_len + len(file_data) > index >= last_len:
                                if last_len + len(file_data) - K < index:
                                    index -= K - ((last_len + len(file_data)) - index)
                                index -= last_len
                                data_point = file_data[index:index + K]
                                data[i].extend(data_point)

                        last_len += len(file_data)
                    except EOFError:
                        break
        return [(x[0], x[1], (x[2][0], x[2][1], x[2][2]), x[3]) for x in zip(*data)]

    def clear_dir(self):
        files = [x for x in os.listdir(self.pickle_dir) if x.endswith(".pkl")]
        for file in files:
            os.remove(f"{self.pickle_dir}/{file}")
        self.processed_count = 0



import copy
import gc
from multiprocessing import Pool

import numpy as np
import torch as th

from mu_alpha_zero.General.memory import GeneralMemoryBuffer
from mu_alpha_zero.General.mz_game import MuZeroGame
from mu_alpha_zero.General.search_tree import SearchTree
from mu_alpha_zero.MuZero.MZ_MCTS.mz_node import MzAlphaZeroNode
from mu_alpha_zero.MuZero.Network.networks import MuZeroNet
from mu_alpha_zero.MuZero.lazy_arrays import LazyArray
from mu_alpha_zero.MuZero.utils import match_action_with_obs, resize_obs, scale_action, scale_reward, scale_state
from mu_alpha_zero.config import MuZeroConfig
from mu_alpha_zero.mem_buffer import MuZeroFrameBuffer


class MuZeroSearchTree(SearchTree):

    def __init__(self, game_manager: MuZeroGame, muzero_config: MuZeroConfig):
        self.game_manager = game_manager
        self.muzero_config = muzero_config
        self.buffer = MuZeroFrameBuffer(self.muzero_config.frame_buffer_size, self.game_manager.get_noop(),
                                        self.muzero_config.net_action_size)
        self.min_max_q = [float("inf"), -float("inf")]

    def play_one_game(self, network_wrapper: MuZeroNet, device: th.device, dir_path: str or None = None) -> list:
        num_steps = self.muzero_config.num_steps
        frame_skip = self.muzero_config.frame_skip
        state = self.game_manager.reset()
        state = resize_obs(state, self.muzero_config.target_resolution)
        state = scale_state(state)
        self.buffer.init_buffer(state)
        data = []
        for step in range(num_steps):
            pi, (v, latent) = self.search(network_wrapper, state, None, device)
            move = self.game_manager.select_move(pi)
            _, pred_v = network_wrapper.prediction_forward(latent.unsqueeze(0), predict=True)
            state, rew, done = self.game_manager.frame_skip_step(move, None, frame_skip=frame_skip)
            rew = scale_reward(rew)
            state = resize_obs(state, self.muzero_config.target_resolution)
            state = scale_state(state)
            if done:
                break
            move = scale_action(move, self.game_manager.get_num_actions())
            frame = self.buffer.concat_frames().detach().cpu().numpy()
            data.append(
                (pi, v, (rew, move, float(pred_v[0])), frame if dir_path is None else LazyArray(frame, dir_path)))
            self.buffer.add_frame(state, move)

        gc.collect()  # To clear any memory leaks, might not be necessary.
        return data

    def search(self, network_wrapper, state: np.ndarray, current_player: int or None, device: th.device,
               tau: float or None = None):
        if len(self.buffer) == 0:
            self.buffer.init_buffer(state)
        num_simulations = self.muzero_config.num_simulations
        if tau is None:
            tau = self.muzero_config.tau

        root_node = MzAlphaZeroNode()
        state_ = network_wrapper.representation_forward(
            self.buffer.concat_frames().permute(2, 0, 1).unsqueeze(0)).squeeze(0)
        pi, v = network_wrapper.prediction_forward(state_.unsqueeze(0), predict=True)
        pi = pi.flatten().tolist()
        root_node.expand_node(state_, pi, 0)
        for simulation in range(num_simulations):
            current_node = root_node
            path = [current_node]
            action = None
            while current_node.was_visited():
                current_node, action = current_node.get_best_child(c=self.muzero_config.c, c2=self.muzero_config.c2)
                path.append(current_node)

            action = scale_action(action, self.game_manager.get_num_actions())

            current_node_state_with_action = match_action_with_obs(current_node.parent().state, action)
            next_state, reward = network_wrapper.dynamics_forward(current_node_state_with_action.unsqueeze(0),
                                                                  predict=True)
            reward = reward[0][0]
            v = self.game_manager.game_result(current_node.current_player)
            if v is None or not v:
                pi, v = network_wrapper.prediction_forward(next_state.unsqueeze(0), predict=True)
                pi = pi.flatten().tolist()
                v = v.flatten().tolist()[0]
                current_node.expand_node(next_state, pi, reward)
            self.backprop(v, path)

        action_probs = root_node.get_self_action_probabilities(tau=tau)
        root_val_latent = (root_node.get_self_value(), root_node.get_latent())
        root_node = None
        return action_probs, root_val_latent

    def backprop(self, v, path):
        G = 0
        gamma = self.muzero_config.gamma
        for node in reversed(path):
            G = v + node.reward if G == 0 else G
            node.total_value += G
            node.update_q(G)
            self.update_min_max_q(node.q)
            node.scale_q(self.min_max_q[0], self.min_max_q[1])
            node.times_visited += 1
            G = node.reward + gamma * G

    def self_play(self, net: MuZeroNet, device: th.device, num_games: int, memory: GeneralMemoryBuffer) -> tuple[
        int, int, int]:
        for game in range(num_games):
            game_results = self.play_one_game(net, device)
            memory.add_list(game_results)

        return None, None, None

    def make_fresh_instance(self):
        return MuZeroSearchTree(self.game_manager.make_fresh_instance(), copy.deepcopy(self.muzero_config))

    def step_root(self, action: int or None):
        # I am never reusing the tree in MuZero.
        pass

    def update_min_max_q(self, q):
        self.min_max_q[0] = min(self.min_max_q[0], q)
        self.min_max_q[1] = max(self.min_max_q[1], q)

    @staticmethod
    def parallel_self_play(nets: list, trees: list, memory: GeneralMemoryBuffer, device: th.device, num_games: int,
                           num_jobs: int):
        with Pool(num_jobs) as p:
            if memory.is_disk and memory.full_disk:
                results = p.starmap(p_self_play, [
                    (nets[i], trees[i], copy.deepcopy(device), num_games // num_jobs, copy.deepcopy(memory),None) for i in
                    range(len(nets))])
            else:
                results = p.starmap(p_self_play,
                                    [(nets[i], trees[i], copy.deepcopy(device), num_games // num_jobs, None,memory.dir_path) for i in
                                     range(len(nets))])
        for result in results:
            memory.add_list(result)

        return None, None, None


def p_self_play(net, tree, dev, num_g, mem, dir_path: str or None = None):
    data = []
    for _ in range(num_g):
        game_results = tree.play_one_game(net, dev, dir_path=dir_path)
        if mem is not None:
            mem.add_list(game_results)
        else:
            data.extend(game_results)
    return data



from mu_alpha_zero.MuZero.muzero import MuZero
from mu_alpha_zero.AlphaZero.alpha_zero import AlphaZero
from mu_alpha_zero.General.memory import GeneralMemoryBuffer
from mu_alpha_zero.General.network import GeneralNetwork
from mu_alpha_zero.General.arena import GeneralArena
from mu_alpha_zero.General.az_game import AlphaZeroGame
from mu_alpha_zero.General.mz_game import MuZeroGame
from mu_alpha_zero.General.search_tree import SearchTree
from mu_alpha_zero.AlphaZero.constants import SAMPLE_AZ_ARGS, SAMPLE_MZ_ARGS
from mu_alpha_zero.General.utils import find_project_root, clear_disk_data
from mu_alpha_zero.Game.tictactoe_game import TicTacToeGameManager
from mu_alpha_zero.Game.asteroids import Asteroids
from mu_alpha_zero.MuZero.Network.networks import MuZeroNet
from mu_alpha_zero.AlphaZero.Network.nnet import AlphaZeroNet
from mu_alpha_zero.MuZero.utils import resize_obs




# --- Linear input sizes for different board and atari sizes ---
# 4608 for (5x5) board
# 512 for (3x3) board
# 32768 for (10x10) board
# 18432 for (8x8) board
# 8192 for (6x6) atari

from dataclasses import dataclass

@dataclass
class Config:
    num_net_channels: int = 512
    num_net_in_channels: int = 1
    net_dropout: float = 0.3
    net_action_size: int = 14
    num_simulations: int = 800
    self_play_games: int = 100
    num_iters: int = 50
    epochs: int = 1600
    lr: float = 1.2166489157239912e-05
    max_buffer_size: int = 70_000
    num_pit_games: int = 40
    random_pit_freq: int = 2
    batch_size: int = 255
    tau: float = 1
    arena_tau: float = 1e-2
    c: float = 1
    checkpoint_dir: str = None
    update_threshold: float = 0.6
    num_workers: int = 5
    log_epsilon: float = 1e-9
    zero_tau_after: int = 5
    az_net_linear_input_size: int = 8192
    log_dir: str = "Logs"
    pushbullet_token: str = None

    def to_dict(self):
        return self.__dict__

    @staticmethod
    def from_args(args: dict):
        missing = []
        config = MuZeroConfig()
        for key, val in args.items():
            if key not in config.to_dict().keys():
                missing.append(key)
            setattr(config, key, val)

        if len(missing) > 0:
            print(f"The following keys were missing from the default config and were added: {missing}.")
        return config


@dataclass
class MuZeroConfig(Config):
    num_net_channels: int = 512
    num_net_out_channels: int = 256
    num_net_in_channels: int = 1
    net_dropout: float = 0.3
    net_action_size: int = 14
    net_latent_size: int = 36
    num_simulations: int = 800
    self_play_games: int = 100
    K: int = 5
    gamma: float = 0.997
    frame_buffer_size: int = 32
    frame_skip: int = 4
    num_steps: int = 400
    num_iters: int = 50
    epochs: int = 1600
    lr: float = 1.2166489157239912e-05
    max_buffer_size: int = 70_000
    num_pit_games: int = 40
    random_pit_freq: int = 2
    batch_size: int = 255
    tau: float = 1
    arena_tau: float = 1e-2
    c: float = 1
    c2: int = 19652
    alpha: float = 0.8
    checkpoint_dir: str = None
    update_threshold: float = 0.6
    num_workers: int = 5
    log_epsilon: float = 1e-9
    zero_tau_after: int = 5
    beta: int = 1
    pickle_dir: str = "Pickles/Data"
    target_resolution: tuple[int, int] = (96, 96)
    az_net_linear_input_size: int = 8192
    log_dir: str = None
    pushbullet_token: str = None
    show_tqdm: bool = False


@dataclass
class AlphaZeroConfig(Config):
    board_size: int = 8
    num_net_channels: int = 512
    num_net_in_channels: int = 1
    net_dropout: float = 0.3
    net_action_size: int = board_size ** 2
    num_simulations: int = 1317
    self_play_games: int = 300
    num_iters: int = 50
    epochs: int = 500
    lr: float = 0.0032485504583772953
    max_buffer_size: int = 100_000
    num_pit_games: int = 40
    random_pit_freq: int = 3
    batch_size: int = 256
    tau: float = 1
    arena_tau: float = 0.04139160592420218
    c: float = 1
    checkpoint_dir: str = None
    update_threshold: float = 0.6
    minimax_depth: int = 4
    show_tqdm: bool = True
    num_workers: int = 5
    num_to_win: int = 5
    log_epsilon: float = 1e-9
    zero_tau_after: int = 5
    az_net_linear_input_size: int = 18432
    log_dir: str = None
    pushbullet_token: str = None



import os
import shutil
import numpy as np
import optuna
# import pygraphviz
import torch as th
from IPython import get_ipython


from mu_alpha_zero.AlphaZero.Network.nnet import AlphaZeroNet
from mu_alpha_zero.AlphaZero.constants import SAMPLE_AZ_ARGS as test_args
from mu_alpha_zero.mem_buffer import MemBuffer
from mu_alpha_zero.config import MuZeroConfig, Config, AlphaZeroConfig



class DotDict(dict):
    def __getattr__(self, name):
        return self[name]

    def __setattr__(self, name, value):
        self[name] = value


def augment_experience_with_symmetries(game_experience: list, board_size) -> list:
    game_experience_ = []
    for state, pi, v, _ in game_experience:
        pi = np.array([x for x in pi.values()])
        game_experience_.append((state, pi, v))
        for axis, k in zip([0, 1], [1, 3]):
            state_ = np.rot90(state.copy(), k=k)
            pi_ = np.rot90(pi.copy().reshape(board_size, board_size), k=k).flatten()
            game_experience_.append((state_, pi_, v))
            del state_, pi_
            state_ = np.flip(state.copy(), axis=axis)
            pi_ = np.flip(pi.copy().reshape(board_size, board_size), axis=axis).flatten()
            game_experience_.append((state_, pi_, v))

    return game_experience_


def rotate_stack(state: np.ndarray, k: int):
    for dim in range(state.shape[0]):
        state[dim] = np.rot90(state[dim], k=k)
    return state


def flip_stack(state: np.ndarray, axis: int):
    for dim in range(state.shape[0]):
        state[dim] = np.flip(state[dim], axis=axis)
    return state


def make_channels(game_experience: list):
    experience = []
    for state, pi, v, current_player, move in game_experience:
        state = make_channels_from_single(state)
        experience.append((state, pi, v, current_player, move))

    return experience


def make_channels_from_single(state: np.ndarray):
    player_one_state = np.where(state == 1, 1, 0)  # fill with 1 where player 1 has a piece else 0
    player_minus_one_state = np.where(state == -1, 1, 0)  # fill with 1 where player -1 has a piece else 0
    empty_state = np.where(state == 0, 1, 0)  # fill with 1 where empty spaces else 0
    return np.stack([state, player_one_state, player_minus_one_state, empty_state], axis=0)


def mask_invalid_actions(probabilities: np.ndarray, observations: np.ndarray, board_size) -> np.ndarray:
    to_print = ""  # for debugging
    mask = np.where(observations != 0, -5, observations)
    mask = np.where(mask == 0, 1, mask)
    mask = np.where(mask == -5, 0, mask)
    valids = probabilities.reshape(-1, board_size ** 2) * mask.reshape(-1, board_size ** 2)
    valids_sum = valids.sum()
    if valids_sum == 0:
        # When no valid moves are available (shouldn't happen) sum of valids is 0, making the returned valids an array
        # of nan's (result of division by zero). In this case, we create a uniform probability distribution.
        to_print += f"Sum of valid probabilities is 0. Creating a uniform probability...\nMask:\n{mask}"
        valids = np.full(valids.shape, 1.0 / np.prod(valids.shape))
    else:
        valids = valids / valids_sum  # normalize

    if len(to_print) > 0:
        print(to_print, file=open("masking_message.txt", "w"))
    return valids


def mask_invalid_actions_batch(states: th.tensor) -> th.tensor:
    masks = []
    for state in states:
        np_state = state.detach().cpu().numpy()
        mask = np.where(np_state != 0, -5, np_state)
        mask = np.where(mask == 0, 1, mask)
        mask = np.where(mask == -5, 0, mask)
        masks.append(mask)

    return th.tensor(np.array(masks), dtype=th.float32).squeeze(1)


def check_args(args: dict):
    required_keys = ["num_net_channels", "num_net_in_channels", "net_dropout", "net_action_size", "num_simulations",
                     "self_play_games", "num_iters", "epochs", "lr", "max_buffer_size", "num_pit_games",
                     "random_pit_freq", "board_size", "batch_size", "tau", "c", "checkpoint_dir", "update_threshold"]

    for key in required_keys:
        if key not in args:
            raise KeyError(f"Missing key {key} in args dict. Please supply all required keys.\n"
                           f"Required keys: {required_keys}.")


def calculate_board_win_positions(n: int, k: int):
    return get_num_horizontal_conv_slides(n, k) * (n * 2 + 2) + 4 * sum(
        [get_num_horizontal_conv_slides(x, k) for x in range(k, n)])


def get_num_horizontal_conv_slides(board_size: int, kernel_size: int) -> int:
    return (board_size - kernel_size) + 1


def az_optuna_parameter_search(n_trials: int, init_net_path: str, storage: str, study_name: str, game, config: AlphaZeroConfig):
    """
    Performs a hyperparameter search using optuna. This method is meant to be called using the start_jobs.py script.
    For this method to work, a mysql database must be running on the storage address and an optuna study with the
    given name and the 'maximize' direction must exist.

    :param n_trials: num of trials to run the search for.
    :param init_net_path: The path to the initial network to use for all trials.
    :param storage: The mysql storage string. Specifies what database to use.
    :param study_name: Name of the study to use.
    :param game: The game instance to use.
    :param config: The config to use for the search.
    :return:
    """

    def objective(trial):
        num_mc_simulations = trial.suggest_int("num_mc_simulations", 60, 1600)
        num_self_play_games = trial.suggest_int("num_self_play_games", 50, 200)
        num_epochs = trial.suggest_int("num_epochs", 100, 400)
        lr = trial.suggest_float("lr", 1e-4, 1e-2, log=True)
        temp = trial.suggest_float("temp", 0.5, 1.5)
        arena_temp = trial.suggest_float("arena_temp", 1e-2, 0.5)
        cpuct = trial.suggest_float("cpuct", 0.5, 5)
        log_epsilon = trial.suggest_float("log_epsilon", 1e-10, 1e-7, log=True)

        trial_config.num_simulations = num_mc_simulations
        trial_config.self_play_games = num_self_play_games
        trial_config.epochs = num_epochs
        trial_config.lr = lr
        trial_config.tau = temp
        trial_config.c = cpuct
        trial_config.arena_tau = arena_temp
        trial_config.num_iters = 5
        trial_config.log_epsilon = log_epsilon
        search_tree = McSearchTree(game.make_fresh_instance(), trial_config)
        trainer = Trainer.from_state_dict(init_net_path, trial_config, game, search_tree)
        print(f"Trial {trial.number} started.")
        trainer.train()
        win_freq = trainer.get_arena_win_frequencies_mean()
        trial.report(win_freq, trial.number)
        print(f"Trial {trial.number} finished with win freq {win_freq}.")
        del trainer
        return win_freq

    from mu_alpha_zero.AlphaZero.Network.trainer import Trainer  # import here to avoid circular imports
    from mu_alpha_zero.AlphaZero.MCTS.az_search_tree import McSearchTree
    trial_config = config
    trial_config.show_tqdm = False
    study = optuna.load_study(study_name=study_name, storage=storage)
    study.optimize(objective, n_trials=n_trials)


def build_net_from_config(muzero_config: Config, device) -> AlphaZeroNet:
    network = AlphaZeroNet(muzero_config.num_net_in_channels, muzero_config.num_net_channels,
                           muzero_config.net_dropout, muzero_config.net_action_size,
                           muzero_config.az_net_linear_input_size)
    return network.to(device)


def build_all_from_config(muzero_alphazero_config: Config, device, lr=None, buffer_size=None) -> tuple[
    AlphaZeroNet, th.optim.Optimizer, MemBuffer]:
    if lr is None:
        lr = muzero_alphazero_config.lr
    if buffer_size is None:
        buffer_size = muzero_alphazero_config.max_buffer_size
    network = build_net_from_config(muzero_alphazero_config, device)
    optimizer = th.optim.Adam(network.parameters(), lr=lr)
    memory = MemBuffer(max_size=buffer_size)
    return network, optimizer, memory


def make_net_from_checkpoint(checkpoint_path: str, args: DotDict | None):
    if args is None:
        args = DotDict(test_args)
    device = th.device("cuda" if th.cuda.is_available() else "cpu")
    net = build_net_from_config(args, device)
    data = th.load(checkpoint_path)
    net.load_state_dict(data["net"])
    return net


def is_notebook():
    try:
        shell = get_ipython().__class__.__name__
        if shell == 'ZMQInteractiveShell':
            return True
        elif shell == 'Shell':
            return True
        elif shell == 'TerminalInteractiveShell':
            return False
        else:
            return False
    except NameError:
        return False


def upload_checkpoint_to_gdrive(files: list, not_notebook_ok: bool = False):
    is_nbt = is_notebook()
    if not is_nbt and not_notebook_ok:
        return
    if not is_nbt:
        raise RuntimeError("This method should only be called from a notebook.")

    for file in files:
        if not os.path.exists("/content/drive/MyDrive/Checkpoints") and os.path.exists("/content/drive/MyDrive"):
            os.mkdir("/content/drive/MyDrive/Checkpoints")
        shutil.copy(file, "/content/drive/MyDrive/Checkpoints")


def visualize_tree(root_node, output_file_name: str, depth_limit: int | None = None):
    graph = pygraphviz.AGraph()
    graph.graph_attr["label"] = "MCTS visualization"
    graph.node_attr["shape"] = "circle"
    graph.edge_attr["color"] = "blue"
    graph.node_attr["color"] = "gold"
    if depth_limit is None:
        depth_limit = float("inf")

    def make_graph(node, parent, g: pygraphviz.AGraph, d_limit: int):

        state_ = None
        if node.state is None:
            state_ = str(np.random.randint(low=0, high=5, size=parent.state.shape))
        else:
            state_ = str(node.state)
        g.add_node(state_)
        if parent != node:
            g.add_edge(str(parent.state), state_)
        if not node.was_visited() or d_limit <= 0:
            return
        # queue_ = deque(root_node.children.values())
        # depth = 1
        # num_children = 25
        # children_iterated = 0
        # parent = root_node

        for child in node.children.values():
            make_graph(child, node, g, d_limit=d_limit - 1 if depth_limit != float("inf") else depth_limit)

    make_graph(root_node, root_node, graph, d_limit=depth_limit)
    graph.layout(prog="dot")
    graph.draw(f"{output_file_name}.png")


def cpp_data_to_memory(data: list, memory: MemBuffer, board_size: int):
    # import pickle
    # test_data = pickle.load(open(f"{find_project_root()}/history.pkl","rb"))
    for game_data in data:
        for state, pi, v in game_data:
            state = th.tensor(state, dtype=th.float32).reshape(board_size, board_size)
            pi = th.tensor(pi, dtype=th.float32)
            memory.add((state, pi, v))

#
# if __name__ == "__main__":
#     cpp_data_to_memory(None,MemBuffer(max_size=10_000),test_args)



import uuid
import atexit
import numpy as np
import os


class LazyArray:
    def __init__(self, array: np.ndarray, directory_path: str):
        self.directory_path = directory_path
        self.path = f"{directory_path}/array_{uuid.uuid4()}.npy"
        np.save(self.path, array)

    def load_array(self):
        return np.load(self.path)

    def __array__(self):
        return self.load_array()

    def remove_array(self):
        os.remove(self.path)



import atexit
from abc import ABC, abstractmethod

import jpype
import numpy as np

from mu_alpha_zero.AlphaZero.Network.nnet import AlphaZeroNet
from mu_alpha_zero.Game.tictactoe_game import TicTacToeGameManager
from mu_alpha_zero.General.az_game import AlphaZeroGame


class Player(ABC):
    """
    To create a custom player, extend this class and implement the choose_move method.
    You can see different implementations below.
    """

    @abstractmethod
    def __init__(self, game_manager: AlphaZeroGame, **kwargs):
        pass

    @abstractmethod
    def choose_move(self, board: np.ndarray, **kwargs) -> tuple[int, int]:
        pass

    @abstractmethod
    def make_fresh_instance(self):
        pass

    def init_kwargs(self, kwargs: dict):
        for key in kwargs.keys():
            setattr(self, key, kwargs[key])


class RandomPlayer(Player):
    def __init__(self, game_manager: TicTacToeGameManager, **kwargs):
        self.game_manager = game_manager
        self.name = self.__class__.__name__
        self.kwargs = kwargs
        self.init_kwargs(kwargs)

    def choose_move(self, board: np.ndarray, **kwargs) -> tuple[int, int]:
        move = self.game_manager.get_random_valid_action(board)
        if "unravel" in kwargs.keys():
            unravel = kwargs["unravel"]
        else:
            unravel = True
        return tuple(move) if unravel else int(move)

    def make_fresh_instance(self):
        return RandomPlayer(self.game_manager.make_fresh_instance(), **self.kwargs)


class NetPlayer(Player):
    def __init__(self, game_manager: TicTacToeGameManager, **kwargs):
        self.game_manager = game_manager
        self.name = self.__class__.__name__
        self.kwargs = kwargs
        self.init_kwargs(kwargs)

    def choose_move(self, board: np.ndarray, **kwargs) -> tuple[int, int]:
        try:
            current_player = kwargs["current_player"]
            device = kwargs["device"]
            tau = kwargs["tau"]
        except KeyError:
            raise KeyError("Missing keyword argument. Please supply kwargs: current_player, device, "
                           "tau")

        pi, _ = self.monte_carlo_tree_search.search(self.network, board, current_player, device, tau=tau)
        move = self.game_manager.select_move(pi)
        self.monte_carlo_tree_search.step_root(None)
        if "unravel" in kwargs.keys():
            unravel = kwargs["unravel"]
        else:
            unravel = True
        return self.game_manager.network_to_board(move) if unravel else move

    def make_fresh_instance(self):
        return NetPlayer(self.game_manager.make_fresh_instance(), **{"network": self.network,
                                                                     "monte_carlo_tree_search": self.monte_carlo_tree_search.make_fresh_instance()})

    def set_network(self, network):
        self.network = network


class TrainingNetPlayer(Player):
    def __init__(self, network: AlphaZeroNet, game_manager: TicTacToeGameManager, args: dict):
        raise NotImplementedError("Don't use this class yet, it produces incorrect results.")
        self.name = self.__class__.__name__
        self.args = self.__init_args(args)
        self.network = network
        self.game_manager = game_manager
        self.traced_path = self.network.trace(self.args["board_size"])

    def __init_args(self, args) -> dict:
        for key in ["checkpoint_dir", "max_depth"]:
            try:
                args.pop(key)
            except KeyError:
                print(f"Key {key} not present.")
        return args

    def choose_move(self, board: np.ndarray, **kwargs) -> tuple[int, int]:
        try:
            current_player = kwargs["current_player"]
            device = kwargs["device"]
            tau = kwargs["tau"]
        except KeyError:
            raise KeyError("Missing keyword argument. Please supply kwargs: current_player, device, "
                           "tau")
        pi = CpSelfPlay.CmctsSearch(board, current_player, tau, self.args, self.traced_path)
        move = self.game_manager.select_move(pi)
        return self.game_manager.network_to_board(move)

    def make_fresh_instance(self):
        raise NotImplementedError


class HumanPlayer(Player):
    def __init__(self, game_manager: TicTacToeGameManager, **kwargs):
        self.name = self.__class__.__name__
        self.game_manager = game_manager
        self.kwargs = kwargs
        self.init_kwargs(kwargs)

    def choose_move(self, board: np.ndarray, **kwargs) -> tuple[int, int]:
        if self.game_manager.headless:
            raise RuntimeError("Cannot play with a human player in headless mode.")
        move = self.game_manager.get_human_input(board)
        return move

    def make_fresh_instance(self):
        return HumanPlayer(self.game_manager.make_fresh_instance(), **self.kwargs)


class JavaMinimaxPlayer(Player):

    def __init__(self, game_manager: AlphaZeroGame, **kwargs):
        jpype.addClassPath(r"C:\Users\Skyr\IdeaProjects\Minimax\build\libs\Minimax-1.0-SNAPSHOT.jar")
        jpype.startJVM()
        self.game_manager = game_manager
        self.name = self.__class__.__name__
        self.kwargs = kwargs
        atexit.register(self.on_shutdown)

    def choose_move(self, board: np.ndarray, **kwargs) -> tuple[int, int]:
        try:
            depth = kwargs["depth"]
            player = kwargs["player"]
        except KeyError:
            raise KeyError("Missing keyword argument. Please supply kwargs: depth, player")
        Minimax = jpype.JClass("dev.skyr.Minimax")
        minimax = Minimax()
        print(self.game_manager.num_to_win)
        ja = jpype.JArray.of(board)
        move = minimax.run(board, player, player, depth, True, self.game_manager.num_to_win)
        return tuple(move)

    def make_fresh_instance(self):
        pass

    def on_shutdown(self):
        jpype.shutdownJVM()


class MinimaxPlayer(Player):
    def __init__(self, game_manager: TicTacToeGameManager, **kwargs):
        # self.evaluate_fn = evaluate_fn
        self.game_manager = game_manager
        self.name = self.__class__.__name__
        self.kwargs = kwargs
        self.init_kwargs(kwargs)

    def choose_move(self, board: np.ndarray, **kwargs) -> tuple[int, int]:
        try:
            depth = kwargs["depth"]
            player = kwargs["player"]
        except KeyError:
            raise KeyError("Missing keyword argument. Please supply kwargs: depth, player")
        move = self.minimax(board.copy(), depth, True, player)[1]
        return tuple(move)

    def minimax(self, board: np.ndarray, depth: int, is_max: bool, player: int, alpha=-float("inf"),
                beta=float("inf")) -> tuple:
        # self.game_manager.check_pg_events()
        eval_ = self.evaluate_fn(board)
        if eval_ is not None:
            return eval_, None
        if depth == 0:
            return self.evaluate_fn(board, check_end=False), None

        if is_max:
            best_score = -float("inf")
            best_move = None
            for move in self.game_manager.get_valid_moves(board):
                board[move[0]][move[1]] = player
                score = self.minimax(board.copy(), depth - 1, False, -player, alpha, beta)[0]
                # print(score)
                board[move[0]][move[1]] = 0
                if score > best_score:
                    best_move = move
                best_score = max(score, best_score)
                alpha = max(alpha, best_score)

                if alpha >= beta:
                    break
            # if best_move is None:
            #     return self.evaluate_fn(board, player), None
            return best_score, best_move
        else:
            best_score = float("inf")
            best_move = None
            for move in self.game_manager.get_valid_moves(board):
                board[move[0]][move[1]] = player
                score = self.minimax(board.copy(), depth - 1, True, -player, alpha, beta)[0]
                if score is None:
                    continue
                # print(score)
                board[move[0]][move[1]] = 0
                best_score = min(score, best_score)
                beta = min(beta, best_score)

                if beta <= alpha:
                    break
            # if best_move is None:
            #     return self.evaluate_fn(board, player), None
            return best_score, best_move

    def make_fresh_instance(self):
        return MinimaxPlayer(self.game_manager.make_fresh_instance(), **self.kwargs)



from setuptools import setup

setup(
    name="mu_alpha_zero_library",
    version="1.0.7",
    description="Library for running and training MuZero and AlphaZero models.",
    author="Skyr",
    install_requires=open("requirements.txt").read().strip().split("\n"),
    long_description=open('README.md',encoding="utf-8").read(),
    long_description_content_type='text/markdown',
    package_data={
        "mu_alpha_zero": ["*.txt", "*.root"]
    },
    url="https://github.com/Skirlax/MuAlphaZeroLibrary",

)



import math
from mu_alpha_zero.AlphaZero.MCTS.az_node import AlphaZeroNode


class MzAlphaZeroNode(AlphaZeroNode):
    def __init__(self, select_probability=0, parent=None, times_visited_init=0):
        super().__init__(0, select_probability, parent, times_visited_init)
        self.reward = 0

    def get_best_child(self, c=1.5, c2=19652):
        best_utc = -float("inf")
        best_child = None
        best_action = None
        for action, child in self.children.items():
            child_utc = child.calculate_utc_score(c=c, c2=c2)
            if child_utc > best_utc:
                best_utc = child_utc
                best_child = child
                best_action = action
        return best_child, best_action

    def get_value_pred(self, prediction_forward: callable):
        return prediction_forward(self.state)

    def expand_node(self, state, action_probabilities, im_reward) -> None:

        self.state = state.clone()
        self.reward = im_reward
        for action, probability in enumerate(action_probabilities):
            node = MzAlphaZeroNode(select_probability=probability, parent=self)
            self.children[action] = node

    def get_immediate_reward(self, dynamics_forward: callable, action: int):
        return dynamics_forward(self.state, action)

    def calculate_utc_score(self, c=1.5, c2=19652):
        parent = self.parent()
        if self.q is None:
            # Inspiration taken from https://github.com/suragnair/alpha-zero-general/blob/master/MCTS.py
            utc = c * self.select_probability * math.sqrt(parent.times_visited + 1e-8)
        else:
            utc1 = self.q + self.select_probability * (
                    (math.sqrt(parent.times_visited)) / (1 + self.times_visited))
            utc2 = c + math.log((parent.times_visited + c2 + 1) / c2)
            utc = utc1 * utc2

        return utc

    def scale_q(self, min_q, max_q):
        if min_q == max_q:
            return
        self.q = (self.q - min_q) / (max_q - min_q)



from torch.multiprocessing import set_start_method

from mu_alpha_zero.General.utils import not_zero

set_start_method('spawn', force=True)
from copy import deepcopy

from typing import Type

import torch as th
from tqdm import tqdm
from mu_alpha_zero.AlphaZero.Arena.arena import Arena
from mu_alpha_zero.AlphaZero.Arena.players import RandomPlayer, Player,NetPlayer
from mu_alpha_zero.AlphaZero.MCTS.az_search_tree import McSearchTree
from mu_alpha_zero.AlphaZero.Network.nnet import AlphaZeroNet
from mu_alpha_zero.AlphaZero.checkpointer import CheckPointer
from mu_alpha_zero.AlphaZero.logger import LoggingMessageTemplates, Logger
from mu_alpha_zero.AlphaZero.utils import build_all_from_config
from mu_alpha_zero.General.arena import GeneralArena
from mu_alpha_zero.General.az_game import AlphaZeroGame
from mu_alpha_zero.General.memory import GeneralMemoryBuffer
from mu_alpha_zero.General.network import GeneralNetwork
from mu_alpha_zero.General.search_tree import SearchTree
from mu_alpha_zero.MuZero.JavaGateway.java_manager import JavaManager
from mu_alpha_zero.config import Config
from mu_alpha_zero.mem_buffer import MemBuffer


# joblib.parallel.BACKENDS['multiprocessing'].use_dill = True


class Trainer:
    def __init__(self, network: GeneralNetwork, game: AlphaZeroGame,
                 optimizer: th.optim, memory: GeneralMemoryBuffer,
                 muzero_alphazero_config: Config, checkpointer: CheckPointer,
                 search_tree: SearchTree, net_player: Player,
                 device, headless: bool = True,
                 opponent_network_override: th.nn.Module or None = None,
                 arena_override: GeneralArena or None = None,
                 java_manager: JavaManager = None) -> None:
        self.muzero_alphazero_config = muzero_alphazero_config
        self.device = device
        self.headless = headless
        self.game_manager = game
        self.mcts = search_tree
        self.net_player = net_player
        self.network = network
        self.opponent_network = self.network.make_fresh_instance() if opponent_network_override is None else opponent_network_override
        self.optimizer = optimizer
        self.memory = memory
        self.java_manager = java_manager
        self.arena = Arena(self.game_manager, self.muzero_alphazero_config,
                           self.device) if arena_override is None else arena_override
        self.checkpointer = checkpointer
        self.logger = Logger(logdir=self.muzero_alphazero_config.log_dir,
                             token=self.muzero_alphazero_config.pushbullet_token)
        self.arena_win_frequencies = []
        self.losses = []

    @classmethod
    def from_checkpoint(cls, net_class: Type[GeneralNetwork], tree_class: Type[SearchTree],
                        net_player_class: Type[Player],
                        checkpoint_path: str, checkpoint_dir: str,
                        game: AlphaZeroGame, headless: bool = True,
                        checkpointer_verbose: bool = False,
                        arena_override: GeneralArena or None = None,
                        mem: GeneralMemoryBuffer or None = None):
        device = th.device("cuda" if th.cuda.is_available() else "cpu")
        checkpointer = CheckPointer(checkpoint_dir, verbose=checkpointer_verbose)

        network_dict, optimizer_dict, memory, lr, args, opponent_dict = checkpointer.load_checkpoint_from_path(
            checkpoint_path)
        conf = Config.from_args(args)
        tree = tree_class(game.make_fresh_instance(), conf)
        if "fc1.weight" in network_dict:
            conf.az_net_linear_input_size = network_dict["fc1.weight"].shape[1]
        network = net_class.make_from_config(conf).to(device)
        opponent_network = network.make_fresh_instance().to(device)
        optimizer = th.optim.Adam(network.parameters(), lr=lr)
        # opponent_network = build_net_from_args(args, device)
        net_player = net_player_class(game.make_fresh_instance(),
                                      **{"network": network, "monte_carlo_tree_search": tree})
        network.load_state_dict(network_dict)
        opponent_network.load_state_dict(opponent_dict)

        try:
            optimizer.load_state_dict(optimizer_dict)
        except ValueError:
            print("Couldn't load optimizer dict.")
        if memory is None:
            memory = mem
        return cls(network, game, optimizer, memory, conf, checkpointer, tree, net_player, device, headless=headless,
                   arena_override=arena_override)

    @classmethod
    def create(cls, muzero_alphazero_config: Config, game: AlphaZeroGame, network: GeneralNetwork,
               search_tree: SearchTree,
               net_player: Player,
               headless: bool = True,
               checkpointer_verbose: bool = False,
               arena_override: GeneralArena or None = None,
               memory_override: GeneralMemoryBuffer or None = None,
               java_manager: JavaManager or None = None):
        device = th.device("cuda" if th.cuda.is_available() else "cpu")
        _, optimizer, mem = build_all_from_config(muzero_alphazero_config, device)
        memory = mem if memory_override is None else memory_override
        checkpointer = CheckPointer(muzero_alphazero_config.checkpoint_dir, verbose=checkpointer_verbose)
        return cls(network, game, optimizer, memory, muzero_alphazero_config, checkpointer, search_tree, net_player,
                   device,
                   headless=headless, arena_override=arena_override, java_manager=java_manager)

    @classmethod
    def from_state_dict(cls, path: str, muzero_alphazero_config: Config, game: AlphaZeroGame, search_tree: SearchTree,
                        headless: bool = True,
                        checkpointer_verbose: bool = False, java_manager: JavaManager or None = None):
        device = th.device("cuda" if th.cuda.is_available() else "cpu")
        net, optimizer, memory = build_all_from_config(muzero_alphazero_config, device)
        checkpointer = CheckPointer(muzero_alphazero_config.checkpoint_dir, verbose=checkpointer_verbose)
        net.load_state_dict(th.load(path))
        net_player = NetPlayer(game.make_fresh_instance(), **{"network": net, "monte_carlo_tree_search": search_tree})
        return cls(net, game, optimizer, memory, muzero_alphazero_config, checkpointer, search_tree, net_player, device,
                   headless=headless,
                   java_manager=java_manager)

    def train(self) -> AlphaZeroNet:
        self.opponent_network.to(self.device)
        self.logger.log(LoggingMessageTemplates.TRAINING_START(self.muzero_alphazero_config.num_iters))
        self.opponent_network.load_state_dict(self.network.state_dict())
        # self.checkpointer.save_state_dict_checkpoint(self.network, "h_search_network")
        num_iters = self.muzero_alphazero_config.num_iters
        epochs = self.muzero_alphazero_config.epochs
        num_simulations = self.muzero_alphazero_config.num_simulations
        self_play_games = self.muzero_alphazero_config.self_play_games
        self.network.eval()
        # self.logger.pushbullet_log(f"Cuda status: {th.cuda.is_available()}")
        for i in self.make_tqdm_bar(range(num_iters), "Training Progress", 0):
            if i >= self.muzero_alphazero_config.zero_tau_after:
                self.muzero_alphazero_config.arena_tau = 0
            with th.no_grad():
                n_jobs = self.muzero_alphazero_config.num_workers
                self.logger.log(LoggingMessageTemplates.SELF_PLAY_START(self_play_games))

                if self.muzero_alphazero_config.num_workers > 1:
                    wins_p1, wins_p2, game_draws = self.mcts.parallel_self_play(self.make_n_networks(n_jobs),
                                                                                self.make_n_trees(n_jobs),
                                                                                self.memory, self.device,
                                                                                self_play_games,
                                                                                n_jobs)
                    if isinstance(self.memory, MemBuffer) and self.memory.is_disk and self.memory.full_disk:
                        self.memory = self.memory.make_fresh_instance()
                else:
                    wins_p1, wins_p2, game_draws = self.mcts.self_play(self.network, self.device, self_play_games,
                                                                       self.memory)
                self.logger.log(LoggingMessageTemplates.SELF_PLAY_END(wins_p1, wins_p2, game_draws, not_zero))
                self.logger.pushbullet_log("Finished self-play.")

            self.checkpointer.save_temp_net_checkpoint(self.network)
            self.logger.log(LoggingMessageTemplates.SAVED("temp checkpoint", self.checkpointer.get_temp_path()))
            self.checkpointer.load_temp_net_checkpoint(self.opponent_network)
            self.logger.log(LoggingMessageTemplates.LOADED("opponent network", self.checkpointer.get_temp_path()))
            self.network.train()
            self.logger.log(LoggingMessageTemplates.NETWORK_TRAINING_START(epochs))
            mean_loss, losses = self.network.train_net(self.memory, self.muzero_alphazero_config)
            self.logger.log(LoggingMessageTemplates.NETWORK_TRAINING_END(mean_loss))
            self.logger.pushbullet_log(f"Mean loss: {mean_loss}, Max loss: {max(losses)}, Min loss: {min(losses)}")
            self.losses.extend(losses)
            self.checkpointer.save_losses(self.losses)
            self.checkpointer.save_checkpoint(self.network, self.opponent_network, self.optimizer,
                                              self.muzero_alphazero_config.lr, i, self.muzero_alphazero_config,
                                              name="latest_trained_net")

            num_games = self.muzero_alphazero_config.num_pit_games
            update_threshold = self.muzero_alphazero_config.update_threshold
            p1_wins, p2_wins, draws, wins_total = self.run_pitting(num_simulations, num_games)
            self.check_model(p1_wins, wins_total, update_threshold, i)
            self.run_pitting_random(num_simulations, num_games, i)

        important_args = {
            "numIters": self.muzero_alphazero_config.num_iters,
            "numSelfPlayGames": self.muzero_alphazero_config.self_play_games,
            "tau": self.muzero_alphazero_config.tau,
            "updateThreshold": self.muzero_alphazero_config.update_threshold,
            "mcSimulations": self.muzero_alphazero_config.num_simulations,
            "c": self.muzero_alphazero_config.c,
            "maxDepth": self.muzero_alphazero_config.max_depth,
            "numPitGames": self.muzero_alphazero_config.num_pit_games
        }

        self.logger.log(LoggingMessageTemplates.TRAINING_END(important_args))
        self.logger.pushbullet_log(LoggingMessageTemplates.TRAINING_END_PSB())
        return self.network

    def make_n_networks(self, n: int) -> list[AlphaZeroNet]:
        """
        Make n identical copies of self.network using deepcopy.

        :param n: The number of copies to make.
        :return: A list of n identical networks.
        """
        return [deepcopy(self.network) for _ in range(n)]

    def make_n_trees(self, n: int) -> list[McSearchTree]:
        """
        Make n new search trees.
        :param n: The number of trees to create.
        :return: A list of n new search trees.
        """
        trees = []
        for i in range(n):
            # manager = self.game_manager.make_fresh_instance()
            tree = self.mcts.make_fresh_instance()
            trees.append(tree)
        return trees

    def get_arena_win_frequencies_mean(self):
        return sum(self.arena_win_frequencies) / not_zero(len(self.arena_win_frequencies))

    def save_latest(self, path):
        state_dict = self.network.state_dict()
        opponent_state_dict = self.opponent_network.state_dict()
        optimizer_state_dict = self.optimizer.state_dict()
        th.save({
            'optimizer': optimizer_state_dict,
            'memory': self.memory,
            'lr': self.muzero_alphazero_config.lr,
            'net': state_dict,
            'opponent_state_dict': opponent_state_dict,
            'args': self.muzero_alphazero_config.to_dict()
        }, path)
        print("Saved latest model data to {}".format(path))

    def make_tqdm_bar(self, iterable, desc, position, leave=True):
        if self.muzero_alphazero_config.show_tqdm:
            return tqdm(iterable, desc=desc, position=position, leave=leave)
        else:
            return iterable

    def get_network(self):
        return self.network

    def get_tree(self):
        return self.mcts

    def get_args(self):
        return self.muzero_alphazero_config

    def run_pitting(self, num_simulations: int, num_games: int):
        self.network.eval()
        self.opponent_network.eval()
        p1 = self.net_player.make_fresh_instance()
        p1.set_network(self.network)
        p2 = self.net_player.make_fresh_instance()
        p2.set_network(self.opponent_network)
        self.logger.log(LoggingMessageTemplates.PITTING_START(p1.name, p2.name, num_games))
        p1_wins, p2_wins, draws = self.arena.pit(p1, p2, num_games, num_mc_simulations=num_simulations,
                                                 one_player=False)
        wins_total = not_zero(p1_wins + p2_wins)
        self.logger.log(LoggingMessageTemplates.PITTING_END(p1.name, p2.name, p1_wins,
                                                            p2_wins, wins_total, draws))
        self.arena_win_frequencies.append(p1_wins / num_games)
        return p1_wins, p2_wins, draws, wins_total

    def run_pitting_random(self, num_simulations: int, num_games: int, i: int):
        if i % self.muzero_alphazero_config.random_pit_freq != 0:
            return
        self.network.eval()
        with th.no_grad():
            random_player = RandomPlayer(self.game_manager.make_fresh_instance(), **{})
            # self.network, self.mcts,
            p1 = self.net_player.make_fresh_instance()
            p1.set_network(self.network)
            self.logger.log(LoggingMessageTemplates.PITTING_START(p1.name, random_player.name, num_games))
            p1_wins_random, p2_wins_random, draws_random = self.arena.pit(p1, random_player, num_games,
                                                                          num_mc_simulations=num_simulations)
            wins_total = not_zero(p1_wins_random + p2_wins_random)
            self.logger.log(
                LoggingMessageTemplates.PITTING_END(p1.name, random_player.name, p1_wins_random,
                                                    p2_wins_random, wins_total, draws_random))

    def check_model(self, p1_wins: int, wins_total: int, update_threshold: float, i: int):
        if p1_wins / wins_total > update_threshold:
            self.logger.log(LoggingMessageTemplates.MODEL_ACCEPT(p1_wins / wins_total,
                                                                 update_threshold))
            self.checkpointer.save_checkpoint(self.network, self.opponent_network, self.optimizer,
                                              self.muzero_alphazero_config.lr, i,
                                              self.muzero_alphazero_config)
            self.logger.log(LoggingMessageTemplates.SAVED("accepted model checkpoint",
                                                          self.checkpointer.get_checkpoint_dir()))
        else:
            self.logger.log(LoggingMessageTemplates.MODEL_REJECT(p1_wins / wins_total,
                                                                 update_threshold))
            self.checkpointer.load_temp_net_checkpoint(self.network)
            self.logger.log(LoggingMessageTemplates.LOADED("previous version checkpoint",
                                                           self.checkpointer.get_temp_path()))
        self.logger.pushbullet_log(LoggingMessageTemplates.ITER_FINISHED_PSB(i))

    def self_play_and_train(self):
        self.network.eval()
        self.mcts.parallel_self_play(self.make_n_networks(self.muzero_alphazero_config.num_workers),
                                     self.make_n_trees(self.muzero_alphazero_config.num_workers),
                                     self.memory, self.device,
                                     self.muzero_alphazero_config.self_play_games,
                                     self.muzero_alphazero_config.num_workers)
        self.network.train()
        mean,_ = self.network.train_net(self.memory,self.muzero_alphazero_config)
        return mean



import atexit
import os
import pickle
import sys

import torch as th

from mu_alpha_zero.AlphaZero.utils import DotDict
from mu_alpha_zero.General.utils import find_project_root
from mu_alpha_zero.config import Config


class CheckPointer:
    def __init__(self, checkpoint_dir: str | None, verbose: bool = True) -> None:
        self.__checkpoint_dir = checkpoint_dir
        self.make_dir()
        self.__checkpoint_num = self.initialize_checkpoint_num()
        self.__name_prefix = "improved_net_"
        self.verbose = verbose
        atexit.register(self.cleanup)

    def make_dir(self) -> None:
        if self.__checkpoint_dir is not None:
            os.makedirs(self.__checkpoint_dir, exist_ok=True)
            return
        root_dir = find_project_root()
        checkpoint_dir = f"{root_dir}/Checkpoints/NetVersions"
        self.__checkpoint_dir = checkpoint_dir
        os.makedirs(checkpoint_dir, exist_ok=True)

    def save_checkpoint(self, net: th.nn.Module, opponent: th.nn.Module, optimizer: th.optim,
                        lr: float,
                        iteration: int, mu_alpha_zero_config: Config, name: str = None) -> None:
        if name is None:
            name = self.__name_prefix + str(self.__checkpoint_num)

        checkpoint_path = f"{self.__checkpoint_dir}/{name}.pth"
        th.save({
            "net": net.state_dict(),
            "optimizer": optimizer.state_dict(),
            "lr": lr,
            "iteration": iteration,
            "args": mu_alpha_zero_config.to_dict(),
            'opponent_state_dict': opponent.state_dict()
        }, checkpoint_path)
        self.print_verbose(f"Saved checkpoint to {checkpoint_path} at iteration {iteration}.")
        self.__checkpoint_num += 1

    def save_state_dict_checkpoint(self, net: th.nn.Module, name: str) -> None:
        checkpoint_path = f"{self.__checkpoint_dir}/{name}.pth"
        th.save(net.state_dict(), checkpoint_path)
        self.print_verbose(f"Saved state dict checkpoint.")

    def load_state_dict_checkpoint(self, net: th.nn.Module, name: str) -> None:
        checkpoint_path = f"{self.__checkpoint_dir}/{name}.pth"
        net.load_state_dict(th.load(checkpoint_path))
        self.print_verbose(f"Loaded state dict checkpoint.")

    def load_checkpoint_from_path(self, checkpoint_path: str) -> tuple:
        sys.modules['mem_buffer'] = sys.modules['mu_alpha_zero.mem_buffer']
        checkpoint = th.load(checkpoint_path)
        self.print_verbose(f"Restoring checkpoint {checkpoint_path} made at iteration {checkpoint['iteration']}.")
        if "memory" in checkpoint:
            memory = checkpoint["memory"]
        else:
            memory = None
        return checkpoint["net"], checkpoint["optimizer"], memory, checkpoint["lr"], \
            DotDict(checkpoint["args"]), checkpoint["opponent_state_dict"]

    def load_checkpoint_from_num(self, checkpoint_num: int) -> tuple:
        checkpoint_path = f"{self.__checkpoint_dir}/{self.__name_prefix}{checkpoint_num}.pth"
        return self.load_checkpoint_from_path(checkpoint_path)

    def clear_checkpoints(self) -> None:
        # This method doesn't obey the verbose flag as it's a destructive operation.

        print("Clearing all checkpoints.")
        answer = input("Are you sure?? (y/n): ")
        if answer != "y":
            print("Aborted.")
            return
        for file_name in os.listdir(self.__checkpoint_dir):
            os.remove(f"{self.__checkpoint_dir}/{file_name}")
        print(f"Cleared {len(os.listdir(self.__checkpoint_dir))} saved checkpoints (all).")

    def save_temp_net_checkpoint(self, net) -> None:
        process_pid = os.getpid()
        os.makedirs(f"{self.__checkpoint_dir}/Temp", exist_ok=True)
        checkpoint_path = f"{self.__checkpoint_dir}/Temp/temp_net_{process_pid}.pth"
        th.save(net.state_dict(), checkpoint_path)

    def load_temp_net_checkpoint(self, net) -> None:
        process_pid = os.getpid()
        checkpoint_path = f"{self.__checkpoint_dir}/Temp/temp_net_{process_pid}.pth"
        net.load_state_dict(th.load(checkpoint_path))

    def initialize_checkpoint_num(self) -> int:
        return len([x for x in os.listdir(self.__checkpoint_dir) if x.endswith(".pth")])

    def get_highest_checkpoint_num(self) -> int:
        return max([int(file_name.split("_")[2].split(".")[0]) for file_name in os.listdir(self.__checkpoint_dir)])

    def get_temp_path(self) -> str:
        return f"{self.__checkpoint_dir}/Temp/temp_net.pth"

    def get_checkpoint_dir(self) -> str:
        return self.__checkpoint_dir

    def get_latest_name_match(self, name: str):
        name_matches = [os.path.join(self.__checkpoint_dir, x) for x in os.listdir(self.__checkpoint_dir) if name in x]
        name_matches.sort(key=lambda x: os.path.getctime(x))
        return name_matches[-1]

    def get_name_prefix(self):
        return self.__name_prefix

    def print_verbose(self, msg: str) -> None:
        if self.verbose:
            print(msg)

    def cleanup(self):
        import shutil
        shutil.rmtree(f"{self.__checkpoint_dir}/Temp/", ignore_errors=True)

    def save_losses(self, losses: list[float]):
        with open(f"{self.__checkpoint_dir}/training_losses.pkl", "wb") as f:
            pickle.dump(losses, f)
